Public Class HeapPriorityQueue
    Private m_NumValues As Long
    Private m_Priorities() As Long
    Private m_Values() As String
    Private m_X() As Integer
    Private m_Y() As Integer

    Private Const WID As Integer = 30
    Private Const HGT As Integer = 15
    Private Const GAP As Integer = 5

    ' Add an item to the queue.
    Public Sub Push(ByVal new_value As String, ByVal new_priority As Long)
        ' Enlarge the value array.
        ReDim Preserve m_Priorities(m_NumValues)
        ReDim Preserve m_Values(m_NumValues)
        ReDim m_X(m_NumValues)
        ReDim m_Y(m_NumValues)

        Dim parent_index As Long
        Dim child_index As Long = m_NumValues

        ' Stop if we reach the root node.
        Do While child_index > 0
            parent_index = (child_index - 1) \ 2

            ' If parent >= new value, stop.
            If m_Priorities(parent_index) >= new_priority Then Exit Do

            ' Move the parent down.
            m_Priorities(child_index) = m_Priorities(parent_index)
            m_Values(child_index) = m_Values(parent_index)

            ' Move up the tree.
            child_index = parent_index
        Loop

        ' Deposit the new value in the parent's old spot.
        m_Priorities(child_index) = new_priority
        m_Values(child_index) = new_value

        m_NumValues += 1
    End Sub

    ' Remove and return the highest priority item.
    Public Function Pop(Optional ByRef priority As Long = 0) As String
        ' Save the return value.
        priority = m_Priorities(0)
        Dim return_value As String = m_Values(0)

        ' Save the last item's value.
        m_NumValues -= 1
        Dim moving_priority As Long = m_Priorities(m_NumValues)
        Dim moving_value As String = m_Values(m_NumValues)

        ' Shrink the arrays.
        If m_NumValues > 0 Then
            ReDim Preserve m_Priorities(m_NumValues - 1)
            ReDim Preserve m_Values(m_NumValues - 1)
            ReDim m_X(m_NumValues - 1)
            ReDim m_Y(m_NumValues - 1)
        End If

        ' Push the new top item down.
        Dim parent_index As Long = 0
        Dim child1_index As Long = parent_index * 2 + 1
        Dim child2_index As Long = parent_index * 2 + 2
        Dim big_child_index As Long
        Do While child1_index < m_NumValues
            ' Find the bigger child value.
            If child2_index >= m_NumValues Then
                ' There's only one child. Use it.
                big_child_index = child1_index
            ElseIf m_Priorities(child1_index) > m_Priorities(child2_index) Then
                ' The first child is bigger.
                big_child_index = child1_index
            Else
                ' The second child is bigger.
                big_child_index = child2_index
            End If

            ' Compare the moving priority to the 
            ' bigger child priority.
            If moving_priority >= m_Priorities(big_child_index) Then
                ' The parent is bigger. We're done.
                Exit Do
            Else
                ' Move the bigger child value up.
                m_Priorities(parent_index) = m_Priorities(big_child_index)
                m_Values(parent_index) = m_Values(big_child_index)
                parent_index = big_child_index
            End If

            ' Calculate the new child indexes.
            child1_index = parent_index * 2 + 1
            child2_index = parent_index * 2 + 2
        Loop

        ' Deposit the new value in the parent's old spot.
        m_Priorities(parent_index) = moving_priority
        m_Values(parent_index) = moving_value

        ' Return the saved value.
        Return return_value
    End Function

    ' Display the heap at the indicated position.
    Public Sub ShowHeap(ByVal back_color As Color, ByVal gr As Graphics, ByVal X As Integer, ByVal Y As Integer, ByVal fnt As Font)
        gr.Clear(back_color)

        ' Position the nodes.
        PositionSubtree(0, X, Y)

        ' Draw lines between the nodes.
        For i As Integer = 0 To m_NumValues \ 2
            DrawChildLines(gr, i)
        Next i

        ' Draw the nodes.
        Dim string_format As New StringFormat
        string_format.Alignment = StringAlignment.Center
        string_format.LineAlignment = StringAlignment.Center
        For i As Integer = 0 To m_NumValues - 1
            DrawNode(gr, i, m_X(i), m_Y(i), fnt, string_format)
        Next i
    End Sub

    ' Draw this node at the indicated position.
    Private Sub DrawNode(ByVal gr As Graphics, ByVal index As Long, ByVal X As Integer, ByVal Y As Integer, ByVal fnt As Font, ByVal string_format As StringFormat)
        Dim rect As New Rectangle(X, Y, WID, HGT)
        Dim rectf As New RectangleF(X, Y, WID, HGT)

        ' Draw the node.
        gr.FillRectangle(Brushes.LightBlue, rect)
        gr.DrawRectangle(Pens.Black, rect)
        gr.DrawString(m_Values(index), fnt, Brushes.Black, rectf, string_format)
    End Sub

    ' Position this subtree.
    Private Sub PositionSubtree(ByVal index As Integer, ByRef X As Integer, ByVal Y As Integer)
        ' Do nothing it this node doesn't exist.
        If index >= m_NumValues Then Exit Sub

        Dim xmin As Integer = X

        ' Position the left subtree.
        If 2 * index + 1 < m_NumValues Then
            PositionSubtree(2 * index + 1, X, Y + HGT + GAP)
        Else
            X += WID / 2
        End If

        ' Position the right subtree.
        If 2 * index + 2 < m_NumValues Then
            PositionSubtree(2 * index + 2, X, Y + HGT + GAP)
        Else
            X += WID / 2
        End If

        ' If there are no children, make room
        ' for this node.
        If 2 * index + 1 >= m_NumValues Then X += WID

        ' Position the node.
        m_X(index) = (xmin + X - WID) / 2
        m_Y(index) = Y
    End Sub

    ' Draw lines from this node to its children.
    Private Sub DrawChildLines(ByVal gr As Graphics, ByVal index As Integer)
        Dim child_index As Integer

        child_index = 2 * index + 1
        If child_index >= m_NumValues Then Exit Sub
        gr.DrawLine(Pens.Black, _
            m_X(index) + WID \ 2, _
            m_Y(index) + HGT \ 2, _
            m_X(child_index) + WID \ 2, _
            m_Y(child_index) + HGT \ 2)

        child_index = 2 * index + 2
        If child_index >= m_NumValues Then Exit Sub
        gr.DrawLine(Pens.Black, _
            m_X(index) + WID \ 2, _
            m_Y(index) + HGT \ 2, _
            m_X(child_index) + WID \ 2, _
            m_Y(child_index) + HGT \ 2)
    End Sub
End Class
